/*
    lwp.d -- Light weight processes.
*/
/*
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 1996, Brian Spilsbury.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/

/*
    Rewritten to use a multiple queue scheme to reduce time, and
    facilitate simulated blocking io, and sleeping without
    busy-looping.
    Changed timing to be millisecond standard.
    Made (sleep) equivelent to (%delay).
    added lwpgetc(), lwpputc(), lwpread(), lwpwrite() to provide
    for transparent blocking character and sequence io. lwpread, and
    lwpwrite aren't used yet.
    Made all streams non-blocking.

    Brian Spilsbury, 1996.
*/

#include "config.h"

#include <sys/timeb.h>
#include <errno.h>

/******************************* EXPORTS ******************************/

lpd main_lpd;
lpd *clwp = &main_lpd;
int critical_level = 0;

pd main_pd;

pd *active = &main_pd;		/* the pd that is attached to clwp */

	/* circular queues, no tails */
pd *running_queue;		/* running pd's  */
pd *blocking_queue;		/* blocking pd's */
pd *delayed_queue;		/* delaying pd's */
pd *dead_queue;			/* dead pd's */
pd *stopped_queue;		/* stopped pd's */
pd *suspended_queue;		/* suspended pd's */
pd *waiting_queue;		/* waiting pd's */

/******************************* IMPORTS ******************************/

extern scheduler_interruption; /* in unixint.c */
extern int writec_PRINTstream();
extern object readc();
extern gc();
extern enum type garbage_parameter;

/******************************* ------- ******************************/

#define thread_switch()  { setTimer(0); enable_scheduler(); \
			     scheduler(0, 0, NULL); }

static int timer_active = FALSE;
static int scheduler_disabled = FALSE;
static int scheduler_level = 0;            /* tito */
static int reset_timer = FALSE;
static int running_processes = 1;
static int awake_processes = 1;
static int absolute_time = 0;
static int housekeeping_time = 0;
static int fd_hightide = 3;		   /* highest fd ever to block + 1 */
static int wake_lowtide = -1;		   /* time to closest sleep-wake */

/* hopefully this will work with DJGPP, but I really have no idea... */
/* Sets for blocking threads, see housekeeping */
static fd_set fd_rd, fd_wr, fd_ex;

object Srunning;
object Ssuspended;
object Swaiting;
object Sstopped;
object Sdead;
object Sblocking;
object Sdelayed;
object siSthread_top_level;

void (*override_redirect_fun)() = NULL;
pd *override_redirect_process = NULL;

static object main_thread;

static
setTimer(long time)
{
  static struct itimerval oldtimer;
  static struct itimerval itimer;

  itimer.it_value.tv_sec = 0;
  itimer.it_value.tv_usec = time;
  itimer.it_interval.tv_sec = 0;
  itimer.it_interval.tv_usec = 0;
  setitimer(ITIMER_REAL, &itimer, &oldtimer);
}

pd *
make_pd(pd *o)
{
  pd *new_pd; lpd *npd;
  
  /* Allocate a new descriptor for the new lwp */
  /* if we already have one, then we have passed it in o... */

  if (o)
    new_pd = o;
  else
    new_pd = (pd *)malloc(sizeof(pd));
  
  /* create a new stack ... */
  if (!o) new_pd->pd_base = (int *)malloc(STACK_SIZE * sizeof(int));

  /* allocate a lisp descriptor:
   * using the calloc here it's possible to avoid the
   * critical section in the various push operations
   */
  if (o)
    npd = new_pd->pd_lpd;
  else
    npd = new_pd->pd_lpd = (lpd *)calloc(1, sizeof(lpd));

  /* initialize it */

  /* bind stack */
  npd->lwp_bds_top = npd->lwp_bind_stack - 1;
  npd->lwp_bds_limit = npd->lwp_bind_stack + BDSSIZE;

  /* C stack */
  /* cssize is different now for the main thread only, but you might
     want to create threads with different stack sizes             */

#ifdef DOWN_STACK
  npd->lwp_cs_org = new_pd->pd_base + STACK_SIZE - 1;
  npd->lwp_cs_limit = new_pd->pd_base - 1;
#else
  npd->lwp_cs_org = new_pd->pd_base;
  npd->lwp_cs_limit = npd->lwp_cs_org + STACK_SIZE;
#endif    
  /* invocation history stack */
  npd->lwp_ihs_top = npd->lwp_ihs_stack - 1;
  npd->lwp_ihs_limit = npd->lwp_ihs_stack + IHSSIZE;
  /* frame stack */
  npd->lwp_frs_top = npd->lwp_frame_stack - 1;
  npd->lwp_frs_limit = npd->lwp_frame_stack + FRSSIZE;
  
  /* constants are fine for a reincarnatee */
  npd->lwp_alloc_temporary = OBJNULL;
  npd->lwp_backq_level = 0;
  npd->lwp_eval1 = 0;
  /* for gc */
  /* we need to rebuild temporary_stream for some reason */
  if (!o) npd->lwp_fmt_temporary_stream = OBJNULL;
  npd->lwp_fmt_temporary_string = OBJNULL;
  
  npd->lwp_PRINTstream = Cnil;
  npd->lwp_PRINTescape = TRUE;
  npd->lwp_PRINTpretty = FALSE;
  npd->lwp_PRINTcircle = FALSE;
  npd->lwp_PRINTbase = 10;
  npd->lwp_PRINTradix = FALSE;
  npd->lwp_PRINTcase = Kdowncase;
  npd->lwp_PRINTgensym = TRUE;
  npd->lwp_PRINTlevel = -1;
  npd->lwp_PRINTlength = -1;
  npd->lwp_PRINTarray = FALSE;
  npd->lwp_write_ch_fun = writec_PRINTstream;
  npd->lwp_output_ch_fun = writec_PRINTstream;
  npd->lwp_read_ch_fun = readc;
  
  npd->lwp_READtable  = symbol_value(Vreadtable);
  npd->lwp_READdefault_float_format = 'S';
  npd->lwp_READbase = 10;
  npd->lwp_READsuppress = FALSE;
  npd->lwp_delimiting_char = OBJNULL;
  npd->lwp_detect_eos_flag = FALSE;
  npd->lwp_in_list_flag = FALSE;
  npd->lwp_dot_flag = FALSE;
  npd->lwp_sharp_eq_context_max = 0;
  
  /* for gc */
  npd->lwp_string_register = OBJNULL;
  npd->lwp_gensym_prefix = OBJNULL;
  npd->lwp_gentemp_prefix = OBJNULL;
  if (!o) npd->lwp_token = OBJNULL;
    
  /* lex_env copy */
  npd->lwp_lex[0] = lex_env[0];
  npd->lwp_lex[1] = lex_env[1];
  npd->lwp_lex[2] = lex_env[2];
  npd->lwp_lex_env = npd->lwp_lex;
  
  /* ihs_push(Cnil) */
  (++npd->lwp_ihs_top)->ihs_function = Cnil;
  npd->lwp_ihs_top->ihs_base = npd->lwp_lex_env;

  /* Now the allocation. If the gc is invoked we are able to mark
   * the objects already allocated
   */

  /* Hmm, this gets more complex with a reincarnatee */
  /* ideally we just want to initialize these destructively */
  /* and hope that this is good enough. */
  /* The main problem is that if other things have been passed these */
  /* and don't expect them to suddenly change, but I'm not sure that this */
  /* can be the case, since these should be local to a thread... */

  if (!o) npd->lwp_fmt_temporary_stream = make_string_output_stream(64);
  /* might need some resetting here? */
  npd->lwp_fmt_temporary_string = npd->lwp_fmt_temporary_stream->sm.sm_object0;
  
  if (!o) npd->lwp_string_register = alloc_simple_string(0);
  if (!o) npd->lwp_gensym_prefix = make_simple_string("G");
  if (!o) npd->lwp_gentemp_prefix = make_simple_string("T");
  if (!o) npd->lwp_token = alloc_simple_string(LISP_PAGESIZE);
  if (!o) npd->lwp_token->st.st_self = alloc_contblock(LISP_PAGESIZE);
  npd->lwp_token->st.st_fillp = 0;
  npd->lwp_token->st.st_hasfillp = TRUE;
  npd->lwp_token->st.st_adjustable = TRUE;

  new_pd->pd_status = SUSPENDED;
  ENQUEUE(new_pd, suspended_queue);	/* needs to be on a queue */
  return new_pd;
}

activate_thread(object thread)
{
  pd *npd = thread->th.th_self;

  /* jump on the new C stack */
  if (sigsetjmp(npd->pd_env, 1) == 0) {
#define STACK_MARGIN 160	/* longjmp writes also under the sp level */
#ifdef DOWN_STACK
# ifdef __linux
    npd->pd_env[0].__jmpbuf[0].__sp =
      stack_align((int)(npd->pd_base) + sizeof(int)*STACK_SIZE - STACK_MARGIN);
    npd->pd_lpd->lwp_cssize =
      npd->pd_env[0].__jmpbuf[0].__sp - (int)npd->pd_base;
# else
    npd->pd_env[JB_SP] =
      stack_align((int)(npd->pd_base) + sizeof(int)*STACK_SIZE - STACK_MARGIN);
    npd->pd_lpd->lwp_cssize =
      npd->pd_env[JB_SP] - (int)npd->pd_base;
# endif
#else
    npd->pd_env[JB_SP] =
      stack_align((int)(npd->pd_base));
    npd->pd_lpd->lwp_cssize = sizeof(int) * STACK_SIZE - STACK_MARGIN;
#endif DOWN_STACK
#ifdef JB_FP
    npd->pd_env[JB_FP] = npd->pd_env[JB_SP];
#endif
    return;
  }

  /* Back here when thread is first resumed.
   *
   * WARNING: args and locals are no more accessible from here on,
   * since we reenter with longjmp onto a new stack.
   */
  
#ifndef sun4sol2
  /* on SunOS sigmask is 0x82001 here */
  sigsetmask(sigblock(0) & ~(sigmask(SIGALRM)));
#endif

/* to get here we've been scheduled */
/* so we aren't in someone else's bind stack */
/* so we should get the defaults for the stdio */

start_critical_section();

for(;;) {	/* mortal coil */
  /* set up local stdio bindings below everything else on the bind stack */
  /* so that they shouldn't be take out of scope ever... */

  bind_var(Vstandard_input, SYM_VAL(Vstandard_input), Cnil);
  bind_var(Vstandard_output, SYM_VAL(Vstandard_output), Cnil);
  bind_var(Verror_output, SYM_VAL(Verror_output), Cnil);
  bind_var(Vquery_io, SYM_VAL(Vquery_io), Cnil);
  bind_var(Vdebug_io, SYM_VAL(Vdebug_io), Cnil);
  bind_var(Vterminal_io, SYM_VAL(Vterminal_io), Cnil);
  bind_var(Vtrace_output, SYM_VAL(Vtrace_output), Cnil);

  { int i;
      for (i = clwp->lwp_nValues; i > 0;)
      VALUES(i) = VALUES(--i);
    VALUES(0) = clwp->lwp_thread->th.th_fun;
    end_critical_section();
    apply(clwp->lwp_nValues+1, siSthread_top_level, &VALUES(0));
    start_critical_section();
  }
  /* Termination */

  {
  pd *tmp = active;

  tmp->pd_status = DEAD;
  running_processes--;
  awake_processes--;
  DEQUEUE(tmp, running_queue);
  ENQUEUE(tmp, dead_queue);
  }

  end_critical_section();
  thread_switch();	/* stack won't have been free'd yet... that's a */
			/* job for housekeeping to think about. */
			/* otherwise dead threads live in limbo waiting */
			/* for reincarnation */
  /* incase we are raised from the dead, we want to do it again */
  start_critical_section();
  }
}

/*
 * switch to the next thread on queue
 */
thread_next()
{
  /* rotate the running-queue */

  /* should devolve into an if, but *shrug* better to guarantee */
  start_critical_section();

  if (override_redirect_process == active) {
	void (*fun)();

	force_resumption(active);

  	if (running_processes > 1) {
    		timer_active = TRUE;
    		setTimer(REALQUANTUM); 
  	} else {
    		timer_active = FALSE;
    		absolute_time = 0;
  		}
	fun = override_redirect_fun;

	override_redirect_fun = NULL;
	override_redirect_process = NULL;

	end_critical_section();
	(*fun)();
	start_critical_section();
	}

  /* unwind the bind stack */
  lwp_bds_unwind(clwp->lwp_bind_stack, clwp->lwp_bds_top);

  ROTQUEUE(running_queue);

  end_critical_section();
  /* housekeeping isn't actually critical
     and turns off the timer while its in there */
  if (absolute_time > housekeeping_time)
	housekeeping();

  /* we need this incase a signal blew us out of the previous housekeeping
     and running_queue is void */

  while (running_queue == NULL) {
    static struct timeb tb;
    ftime(&tb);			/* not sure how portable */
    absolute_time = tb.millitm + tb.time*1000;
    housekeeping();
  }
  start_critical_section();

  /* switch clwp */
  clwp = running_queue->pd_lpd;
  active = running_queue;

  /* restore Values pointer */
  Values = clwp->lwp_Values;

  /* wind the bind stack */
  lwp_bds_wind(clwp->lwp_bind_stack, clwp->lwp_bds_top);

  /* reset the timer */
  if (running_processes > 1) {
    timer_active = TRUE;
    setTimer(REALQUANTUM); 
  } else {
    timer_active = FALSE;
    absolute_time = 0;
  }

  end_critical_section();
  siglongjmp(active->pd_env, 1); 
}

/*
 * Called when time slice expires or explicitily to switch thread
 * New version...
 */
scheduler(int sig, int code, struct sigcontext *scp)
{
  int val;
  static struct timeb tb;

#if defined(SYSV) || defined(__svr4__) || defined(__linux)
  signal(SIGALRM, scheduler);
#endif SYSV

  ftime(&tb);			/* not sure how portable */
  absolute_time = tb.millitm + tb.time*1000;

  if (critical_level > 0) {	/* within critical section */
    scheduler_interrupted = TRUE;
    scheduler_interruption = SCHEDULER_INT;
    return;
  }
  if (scheduler_level > 0) {	/* abilitation  check */
    scheduler_interrupted = TRUE;
    return;
  }

  val = sigsetjmp(active->pd_env, 1);

  if (val == 1)			/* resume interrupted thread execution */
    return;			/* coming back from longjmp in thread_next */

  if (val == 2)			/* coming back from longjmp in GC */
    gc(garbage_parameter);	/* GC will return to the previous thread */

  thread_next();
}

/* TODO: Add in waiting thread condition resolution */
housekeeping()
{
  static struct timeval timeout;
  static pd *p, *q;
  static int tide;
  static fd_set rd, wr, ex;
  /* see if we are polling or lurking */

  /* turn off that bloody timer... */
  setTimer(0);

  if ((running_processes > 1) && (awake_processes > 0)) {
    /* poll */
    /* set timeout to instant */
    timeout.tv_sec = timeout.tv_usec = 0;
    tide = 1;
  } else {
    /* is recovery possible? */

    if (	(running_queue == NULL) &&
	(blocking_queue == NULL) &&
	(delayed_queue == NULL)) {
      /* in a coma... can't awaken itself... */
      /* there is a possibility that a signal */
      /* will, but um, for now assume dead and buried */
      exit(0);			/* bail w/out error */
    }

    /* ok, in theory we can wake up.... so lurk */
    /* set timeout to the shortest sleep resumption time */
    /* if there isn't a sleep resumption time */
    /* block forever... */

    if (wake_lowtide != -1) {
      if (absolute_time >= wake_lowtide) {
	timeout.tv_sec = timeout.tv_usec = 0;
	tide = 1;
      } else {
	timeout.tv_sec = (wake_lowtide-absolute_time)/1000;
	timeout.tv_usec = ((wake_lowtide-absolute_time)%1000)*1000;
	tide = 1;
      }
    } else 	tide = 0;
  }

  memcpy(&rd, &fd_rd, sizeof(fd_set));
  memcpy(&wr, &fd_wr, sizeof(fd_set));
  memcpy(&ex, &fd_ex, sizeof(fd_set));

  /* If there is an error, the sets are undefined, just bail out... */
  /* we'll catch it next time, it was probably a signal interrupting */
  /* us. */
  if (select(fd_hightide, &rd, &wr, &ex, (tide ? &timeout : NULL)) == -1) {
    /* we broke for some reason */
    /* a signal handler may have been invoked... */
    /* someone may have been woken up */
    /* so schedule another housekeeping apointment */
    /* and bail */
    goto out;
  }
	
  /* check for awakened threads */

  if (p = blocking_queue)
    do {
      q = p->pd_next;
      switch(p->pd_fp_mode) {
      case PD_INPUT:
	if (FD_ISSET(fileno(p->pd_fp), &rd)) {
	  FD_CLR(fileno(p->pd_fp), &fd_rd);
	  DEQUEUE(p, blocking_queue);
	  ENQUEUE(p, running_queue);
	  p->pd_status = RUNNING;
	  awake_processes++;
	  if (blocking_queue == NULL) goto endblk;
	}
	break;
      case PD_OUTPUT:
	if (FD_ISSET(fileno(p->pd_fp), &wr)) {
	  FD_CLR(fileno(p->pd_fp), &fd_wr);
	  DEQUEUE(p, blocking_queue);
	  ENQUEUE(p, running_queue);
	  p->pd_status = RUNNING;
	  awake_processes++;
	  if (blocking_queue == NULL) goto endblk;
	}
	break;
      case PD_EXCEPTION:
	if (FD_ISSET(fileno(p->pd_fp), &ex)) {
	  FD_CLR(fileno(p->pd_fp), &fd_ex);
	  DEQUEUE(p, blocking_queue);
	  ENQUEUE(p, running_queue);
	  p->pd_status = RUNNING;
	  awake_processes++;
	  if (blocking_queue == NULL) goto endblk;
	}
	break;
      }
      p = q;
    } while(p != blocking_queue);

  /* if sleeping, check for wakeup.... */

 endblk:	tide = -1;

  /*putchar('.'); fflush(stdout);*/

  if ((wake_lowtide != -1) && ((p = delayed_queue) != NULL))
    do {
      q = p->pd_next;
      if (absolute_time >= p->pd_slice) {
	DEQUEUE(p, delayed_queue);
	ENQUEUE(p, running_queue);
	p->pd_status = RUNNING;
	awake_processes++;
	if (delayed_queue == NULL) break;
      } else {
	/* get new low-tide */
	if ((tide == -1) || (p->pd_slice < tide))
	  tide = p->pd_slice;
      }

      p = q;
    } while(p != delayed_queue);

  wake_lowtide = tide;

  /* requeue waiting threads for resolution checking */
  /* not a good solution, but the only one that I can see */
  /* at least now we will only check them on housekeeping intervals */
  /* just get the timeout cases to set the wake_lowtide on entry */
  /* to prevent oversleeping. */

  while (p = waiting_queue) {
    DEQUEUE(p, waiting_queue);
    ENQUEUE(p, running_queue);
    p->pd_status = RUNNING;
    awake_processes++;
  }

 out:	housekeeping_time = absolute_time + 1000; /* one second */
}

/* this will mostly be used in conjunction with override_redirect */
/* which should be set before entry here. */

force_resumption(pd *p)
{
	start_critical_section();

	switch(p->pd_status) {
		case RUNNING: DEQUEUE(p, running_queue);
				break;
		case STOPPED: DEQUEUE(p, stopped_queue);
				running_processes++;
				awake_processes++;
				break;
		case SUSPENDED: DEQUEUE(p, stopped_queue);
				running_processes++;
				awake_processes++;
				break;
		case DEAD: /* hmmm... raising the dead might be dangerous */
				running_processes++;
				awake_processes++;
				DEQUEUE(p, dead_queue);
				break;
		case WAITING:	DEQUEUE(p, waiting_queue);
				awake_processes++;
				break;
		case DELAYED:	DEQUEUE(p, delayed_queue);
				awake_processes++;
				break;
			/* don't worry about the magic, at worst this can */
				running_processes++;
			/* cause housekeeping to make a false check */
		case BLOCKED:	DEQUEUE(p, blocking_queue);
				/* this needs some fixing */
				switch(p->pd_fp_mode) {
					case PD_INPUT:
						FD_CLR(fileno(p->pd_fp), &fd_rd);
						break;
					case PD_OUTPUT:
						FD_CLR(fileno(p->pd_fp), &fd_wr);
						break;
					case PD_EXCEPTION:
						FD_CLR(fileno(p->pd_fp), &fd_ex);
						break;
					}
				awake_processes++;
				break;
		}

	ENQUEUE(p, running_queue);
	end_critical_section();
}

#ifdef 0
scheduler(int sig, int code, struct sigcontext *scp)
{
  int val;

#if defined(SYSV) || defined(__svr4__) || defined(__linux)
  signal(SIGALRM, scheduler);
#endif SYSV

  absolute_time++;
  if (critical_level > 0) {	/* within critical section */
    scheduler_interrupted = TRUE;
    scheduler_interruption = SCHEDULER_INT;
    return;
  }
  if (scheduler_level > 0) {	/* abilitation  check */
    scheduler_interrupted = TRUE;
    return;
  }

  val = sigsetjmp(running_dead->pd_env, 1);

  if (val == 1)			/* resume interrupted thread execution */
    return;			/* coming back from longjmp in thread_next */

  if (val == 2)			/* coming back from longjmp in GC */
    gc(garbage_parameter);	/* GC will return to the previous thread */

  ROTQUEUE();
  thread_next();
}
#endif /* old version */

/*
 * Handle signal received within critical section
 */
interruption_handler()
{
  scheduler_interrupted = FALSE;

  switch (scheduler_interruption) {

  case SCHEDULER_INT:
    thread_switch();
    break;

  case ERROR_INT:
    sigint();
    break;
  }
}

lwp_bds_wind(bds_ptr base, bds_ptr top)
{
  object temp;

  for (; top >= base; base++) {
    temp = SYM_VAL(base->bds_sym);
    SYM_VAL(base->bds_sym) = base->bds_val;
    base->bds_val = temp;
  }
}

lwp_bds_unwind(bds_ptr base, bds_ptr top)
{
  object temp;

  for (; top >= base; top--) {
    temp = SYM_VAL(top->bds_sym);
    SYM_VAL(top->bds_sym) = top->bds_val;
    top->bds_val = temp;
  }
}

resume(pd *rpd)
{
  register pd *p;

  start_critical_section();

  if (rpd->pd_status == STOPPED) {
  	DEQUEUE(rpd, stopped_queue);
  	running_processes++;
  	awake_processes++;
  } else if (rpd->pd_status == DELAYED) {
  	DEQUEUE(rpd, delayed_queue);
	/* TODO: look for interaction with housekeeping problem for this case... */
  	awake_processes++;
  } else if (rpd->pd_status == SUSPENDED) {
  	DEQUEUE(rpd, suspended_queue);
  	running_processes++;
  	awake_processes++;
	}

  /* else, should we be here? no. hmmm. */
  /* where are the fucking arguments? */

  rpd->pd_status = RUNNING;

  ENQUEUE(rpd, running_queue);

  end_critical_section();

  if (!timer_active)  {
    timer_active = TRUE;
    setTimer(REALQUANTUM);
    }
}

/***********
 *
 * THREADS
 *
 ***********/


siLthread_break_in(int narg)
{
  alarm(0);
  VALUES(0) = Cnil;
  RETURN(1);
}

/* Hmmmmm, what the hell is this supposed to do? */
/* and why bother? */
/* Put this in the TODO basket, as something of dubiousness */

#ifdef 0
siLthread_break_quit(int narg)
{
  /* reset everything in MT */
  pd *p;

  /* this is done in any case to remedy the problem with C-c handling */
  signal(SIGALRM, scheduler);

  if (timer_active) {
     /* reset the critical and disable-scheduler environment    */
      scheduler_disabled = FALSE;
      scheduler_level = 0;
      critical_level = 0;
      scheduler_interrupted = 0;

      /*for (p = running_dead; (p != NULL); p = p->pd_next)*/
      p = running_queue;
      do {
	if (p != &main_pd) {
	  p->pd_status = DEAD;
	} else {
	  p->pd_status = RUNNING;
	  p->pd_thread->th.th_cont = OBJNULL;
	  }
	p = p->pd_next;
      } while(running_queue->pd_next != running_queue);

      if (running_queue != &main_pd) {
	update_queue();
	thread_dext(); 
	/*  here one should deallocate the main-thread function     */
      }
      else
	thread_switch();
    }
  VALUES(0) = Cnil;
  RETURN(1);
}
#endif

siLthread_break_resume(int narg)
{
  /* Restart the timer that might have been
   * changed by the interrupt handling
   */
  signal(SIGALRM, scheduler);  
  if (timer_active)
    thread_switch();
  VALUES(0) = Cnil;
  RETURN(1);
}

Lthread_list(int narg)
{
  pd *p;
  object tmp;
  object tmp2 = CONS(running_queue->pd_thread, Srunning);
  object x = CONS(tmp2, Cnil);

  tmp = x;

  start_critical_section();

  p = running_queue->pd_next;
  while(p != running_queue) {
	tmp2 = CONS(p->pd_thread, Srunning);
	CDR(tmp) = CONS(tmp2, Cnil);
	tmp = CDR(tmp);
	p = p->pd_next;
	}

  if (p = blocking_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Sblocking);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != blocking_queue);

  if (p = delayed_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Sdelayed);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != delayed_queue);

	/* TODO: Should this queue be listed? */
  if (p = dead_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Sdead);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != dead_queue);

  if (p = stopped_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Sstopped);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != stopped_queue);
	
  if (p = suspended_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Ssuspended);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != suspended_queue);

  if (p = waiting_queue)
  	do {
		tmp2 = CONS(p->pd_thread, Swaiting);
		CDR(tmp) = CONS(tmp2, Cnil);
		p = p->pd_next;
		tmp = CDR(tmp);
  	} while(p != waiting_queue);

  end_critical_section();

  VALUES(0) = x;
  RETURN(1);
}

Lmake_thread(int narg, object fun)
{
  object x;
  pd *npd;
  /* Just one argument for the time being */
  check_arg(1);

  if (type_of(fun) == t_symbol) {
    if (fun->s.s_sfdef != NOT_SPECIAL ||
	fun->s.s_mflag)
      FEinvalid_function(fun);
    if (SYM_FUN(fun) == OBJNULL)
      FEundefined_function(fun);
    /*  fun = SYM_FUN(fun); confusing */
  }

start_critical_section();
  /* see if there is a lost soul waiting for reincarnation */
  if (dead_queue) {
	/* ok, lets juice it up */

	npd = dead_queue;
	DEQUEUE(npd, dead_queue);

  	x = npd->pd_lpd->lwp_thread;
	/* enqueued in make_pd */

	x->th.th_fun = fun;
	x->th.th_size = sizeof(pd);
	x->th.th_self = make_pd(npd);	/* reinitialize it */
	x->th.th_cont = OBJNULL;
  } else {
	/* ok, no lost souls, better build a new one */
  	x = alloc_object(t_thread);
  	x->th.th_fun = fun;
  	x->th.th_size = sizeof(pd);
  	x->th.th_self = npd = make_pd(0);
  	x->th.th_cont = OBJNULL;

  	npd->pd_thread = x;
  	npd->pd_slice = 0;

  	/* Backpointer to thread */
  	npd->pd_lpd->lwp_thread = x;

  activate_thread(x);
  }

  /* note: this is created as a suspended thread, and in that queue */

end_critical_section();
  VALUES(0) = x;
  RETURN(1);
}

Ldeactivate(int narg, object thread)
{
  check_arg(1);
  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status != RUNNING)
    FEerror("Cannot deactivate a thread not running", 0);

  start_critical_section();	/* tito */
  thread->th.th_self->pd_status = STOPPED;
  running_processes--;
  awake_processes--;
  if (thread->th.th_self == active) {
  	DEQUEUE(thread->th.th_self, running_queue);
  	ENQUEUE(thread->th.th_self, stopped_queue);
    	critical_level--; /* end_critical_section() */
    	thread_switch();
  } else {
  	DEQUEUE(thread->th.th_self, running_queue);
  	ENQUEUE(thread->th.th_self, stopped_queue);
    	end_critical_section();
	}
  VALUES(0) = Cnil;
  RETURN(1);
}

Lreactivate(int narg, object thread)
{
  check_arg(1);
  start_critical_section();

  if (type_of(thread) != t_thread) {
    FEwrong_type_argument(Sthread, thread);
  }

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status != STOPPED)
    FEerror("Cannot reactivate a thread not stopped", 0);

  start_critical_section();	/* tito */
  thread->th.th_self->pd_status = RUNNING;
  running_processes++;
  awake_processes++;

  if (!timer_active) {
     timer_active = TRUE;
     setTimer(REALQUANTUM);
   }

  end_critical_section();
  VALUES(0) = Cnil;
  RETURN(1);

}

Lkill_thread(int narg, object thread)
{
  pd *tmp;

  check_arg(1);

  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self != NULL) {
    start_critical_section();
    thread->th.th_self->pd_status = DEAD;
    if (thread->th.th_self->pd_lpd == clwp) {
       /* if a thread kills itself the scheduler is to be called */
       tmp = thread->th.th_self;
       DEQUEUE(tmp, running_queue);
       ENQUEUE(tmp, dead_queue);
       critical_level--; /*  end_critical_section() */
       thread_switch();
     }
     else {
       /*thread->th.th_self = NULL;*/
       tmp = thread->th.th_self;
       DEQUEUE(tmp, running_queue);
       ENQUEUE(tmp, dead_queue);
       end_critical_section();
     }
  }
  VALUES(0) = Cnil;
  RETURN(1);
}

Lcurrent_thread(int narg)
{
  check_arg(0);
  VALUES(0) = clwp->lwp_thread;
  RETURN(1);
}

Lthread_status(int narg, object thread)
{
  check_arg(1);

  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self != NULL)
    switch (thread->th.th_self->pd_status) {
    case RUNNING:
      VALUES(0) = Srunning;
      break;
    case SUSPENDED:
      VALUES(0) = Ssuspended;
      break;
    case WAITING:
      VALUES(0) = Swaiting;
      break;
    case STOPPED:
      VALUES(0) = Sstopped;
      break;
    case DEAD:
      VALUES(0) = Sdead;
      break;
    case BLOCKED:
      VALUES(0) = Sblocking;
      break;
    case DELAYED:
      VALUES(0) = Sdelayed;
    default:
      FEerror("Unexpected type for thread ~A", 1, thread);
    }
  else
    VALUES(0) = Sdead;

  RETURN(1);
}


/***************
 *
 * CONTINUATIONS
 *
 ***************/

Lmake_continuation(int narg, object thread)
{
  object x;
  check_arg(1);


  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_cont)
    FEerror("A continuation for thread ~A already exists.", 1, thread);

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status == DEAD) {
    FEerror("Thread ~A is DEAD.", 1, thread);
  }

  x = alloc_object(t_cont);

  x->cn.cn_thread = thread;
  x->cn.cn_resumed = FALSE;
  x->cn.cn_timed_out = FALSE;

  VALUES(0) = thread->th.th_cont = x;
  RETURN(1);
}

/* Returns the thread associated to a continuation */
Lthread_of(int narg, object cont)
{
  check_arg(1);
  if (type_of(cont) != t_cont)
    FEwrong_type_argument(Scont, cont);
  VALUES(0) = cont->cn.cn_thread;
  RETURN(1);
}

/* Returns the continuation associated to a thread, if it exists */
Lcontinuation_of(int narg, object thread)
{
  check_arg(1);
  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);
  if (thread->th.th_cont)
    VALUES(0) = thread->th.th_cont;
  else 
    VALUES(0) = Cnil;
  RETURN(1);
}

Lresume(int narg, object cont, ...)
{ int i;
  va_list args;
  object *thread_Values;

  if (narg < 1) FEtoo_few_arguments(&narg);

  if (Null(cont)) {
    VALUES(0) = Cnil;
    RETURN(1);
  }

  if (type_of(cont) != t_cont)
    FEwrong_type_argument(Scont, cont);

  if (cont->cn.cn_resumed)
    FEerror("The continuation has already been resumed.", 0);
  
  if (cont->cn.cn_timed_out)
    FEerror("The continuation has been timed out.", 0);

  if (cont->cn.cn_thread->th.th_self == NULL)
    FEerror("The continuation cannot be resumed. Its thread is DEAD.", 0);

  if (cont->cn.cn_thread->th.th_self->pd_status != SUSPENDED &&
      cont->cn.cn_thread->th.th_self->pd_status != DELAYED)
    FEerror("The continuation cannot be resumed. Its thread isn't suspended", 0);

  /* Push the arguments on the value stack of thread */

  thread_Values = cont->cn.cn_thread->th.th_self->pd_lpd->lwp_Values;

  va_start(args, cont);
  for (i = 1; i < narg; i++)
    *(thread_Values++) = va_arg(args, object);
  cont->cn.cn_thread->th.th_self->pd_lpd->lwp_nValues = narg-1;

  cont->cn.cn_resumed = TRUE;
  cont->cn.cn_thread->th.th_cont = OBJNULL;

  /* If you are waiting on a slice expiring I reset your slice    */
  cont->cn.cn_thread->th.th_self->pd_slice = 0;

  resume(cont->cn.cn_thread->th.th_self);

  VALUES(0) = cont->cn.cn_thread;
  RETURN(1);
}


/***************
 *
 * SCHEDULING
 *
 ***************/


Ldisable_scheduler(int narg)
{
  check_arg(0);

  scheduler_level++;
  VALUES(0) = Cnil;
  RETURN(1);
}

Lenable_scheduler(int narg)
{
  check_arg(0);

  enable_scheduler();
  VALUES(0) = Cnil;
  RETURN(1);
}

enable_scheduler()
{
   scheduler_level = 0;

   if (scheduler_interrupted) {
     scheduler_interrupted = FALSE;
     thread_switch();
   }
}

Lsuspend(int narg)
{
  check_arg(0);

  if (timer_active) {
       pd *tmp = active;
       tmp->pd_status = SUSPENDED;
       DEQUEUE(tmp, running_queue);
       ENQUEUE(tmp, suspended_queue);
       running_processes--;
       awake_processes--;
       thread_switch();
       /* When resumed it will be provided with the Values to return */
       RETURN(tmp->pd_lpd->lwp_nValues);
     }
  else 
    FEerror("No other active thread.", 0);
}

void
lwpblockon(pd *who, FILE *fp, int mode)
{
	who->pd_fp = fp;
	who->pd_fp_mode = mode;	/* in, out, execept */
	who->pd_status = BLOCKED;

	start_critical_section();

	DEQUEUE(who, running_queue);
	ENQUEUE(who, blocking_queue);
	awake_processes--;

	if (fd_hightide <= fileno(fp))
		fd_hightide = fileno(fp)+1;

	switch(mode) {
		case PD_INPUT:
			FD_SET(fileno(fp), &fd_rd);
			break;
		case PD_OUTPUT:
			FD_SET(fileno(fp), &fd_wr);
			break;
		case PD_EXCEPTION:
			FD_SET(fileno(fp), &fd_ex);
			break;
		}

	end_critical_section();

	thread_switch();
}

int inline
lwpgetc(FILE *fp)
{
	int c;

loop:	errno = 0;
	c = getc(fp);
	if (errno) {
	  lwpblockon(active, fp, PD_INPUT);
	  clearerr(fp);
	  goto loop;
	}
	return(c);
}

void inline
lwpputc(char c, FILE *fp)
{
loop:	errno = 0;
	putc(c, fp);
	if (errno) {
	  lwpblockon(active, fp, PD_OUTPUT);
	  clearerr(fp);
	  goto loop;
	}
	return;
}

int inline
lwpread(char *buf, int len, FILE *fp)
{
	int ind = 0, left = len, n;

loop:	errno = 0;
	n = read(&buf[ind], left, fileno(fp));
	if (errno) {
	  ind += n;
	  left -= n;
	  lwpblockon(active, fp, PD_INPUT);
	  clearerr(fp);
	  goto loop;
	}
	return(ind+n);
}

int inline
lwpwrite(char *buf, int len, FILE *fp)
{
	int ind = 0, left = len, n;

loop:	errno = 0;
	n = write(&buf[ind], left, fileno(fp));
	if (errno) {
	  ind += n;
	  left -= n;
	  lwpblockon(active, fp, PD_OUTPUT);
	  clearerr(fp);
	  goto loop;
	}
	return(ind+n);
}

Ldelay(int narg, object interval)
{ int z;
  
  check_arg(1);
  check_type_non_negative_integer(&interval);
  z = fix(interval);

  if (timer_active) {
    pd *tmp = active;	/* remember who we are */
    lwpsleep(z*1000);	/* lwpsleep is in milliseconds */
    /* When resumed it will be provided with the Values to return */
    RETURN(tmp->pd_lpd->lwp_nValues);
  }
  else
  {
    sleep(z);
  }
}

/* Sleep for at least ms milliseconds */
lwpsleep(int ms)
{
  
  if (timer_active) {
    pd *tmp = active;

    start_critical_section();
    tmp->pd_status = DELAYED;
    DEQUEUE(tmp, running_queue);
    ENQUEUE(tmp, delayed_queue);
    awake_processes--;
       
    tmp->pd_slice = ms + absolute_time;

    if ((wake_lowtide == -1) || (wake_lowtide > tmp->pd_slice))
	wake_lowtide = tmp->pd_slice;

    end_critical_section();

    thread_switch();
  }
  else usleep(ms*1000);	/* milli->micro */
}

/* TODO: Find a way to move this functionality into housekeeping()
   sigh */

Lthread_wait(int narg, object fun, ...)
{ int nr;
  pd *tmp = active;
  va_list args;
  va_start(args, fun);
  
  if (narg < 1) FEtoo_few_arguments(&narg);

  for (;;) {
    
    nr = apply(narg-1, fun, args);
    
    if (VALUES(0) != Cnil)
      break;

    start_critical_section();
    tmp->pd_status = WAITING;

    DEQUEUE(tmp, running_queue);
    ENQUEUE(tmp, waiting_queue);

    awake_processes--;

    end_critical_section();
    thread_switch();
  }

  RETURN(nr);
}
  
/* TODO: whack this into housekeeping() */

Lthread_wait_with_timeout(int narg, object timeout, object fun, ...)
{
  int nr;
  pd *tmp = active;
  va_list args;
  va_start(args, fun);

  if (narg < 2) FEtoo_few_arguments(&narg);
  check_type_non_negative_integer(&timeout);

  /* We have to translate seconds into milliseconds into the future */
  tmp->pd_slice = fix(timeout) * 1000 + absolute_time;
  
  
  for (;;) {
    
    if (absolute_time > tmp->pd_slice) {
      /* the time slice has expired */
      VALUES(0) = Cnil;
      nr = 1;
      break;
    }

    nr = apply(narg-1, fun, args);

    if (VALUES(0) != Cnil)
      break;
    else {
      /* the time slice has not been used */
      absolute_time--;
      thread_switch();
    }

    tmp->pd_status = WAITING;
    DEQUEUE(tmp, running_queue);
    ENQUEUE(tmp, waiting_queue);
    awake_processes--;

    if ((wake_lowtide == -1) || (wake_lowtide > tmp->pd_slice))
	wake_lowtide = tmp->pd_slice;

    thread_switch();
    }
  
  tmp->pd_slice = 0;
  RETURN(nr);
}

enable_lwp()
{
  /* called when we start after a dump */
  FD_ZERO(&fd_rd);
  FD_ZERO(&fd_wr);
  FD_ZERO(&fd_ex);
  signal(SIGALRM, scheduler);
}

init_lwp()
{ pd *temp_pd;

  temp_pd = &main_pd;
  /*PUSH(temp_pd);*/

  FD_ZERO(&fd_rd);
  FD_ZERO(&fd_wr);
  FD_ZERO(&fd_ex);

  ENQUEUE(temp_pd, running_queue);

  main_thread = alloc_object(t_thread);
  main_pd.pd_thread = main_thread;
  main_thread->th.th_fun = Cnil;
  main_thread->th.th_size = sizeof (pd);
  main_thread->th.th_self = &main_pd;
  main_thread->th.th_cont = OBJNULL;
  /* Backpointer to thread */
  main_pd.pd_status = RUNNING;
  main_pd.pd_lpd = &main_lpd;
  main_lpd.lwp_thread = main_thread;
  register_root(&main_thread);

  Srunning = make_ordinary("RUNNING");
  Ssuspended = make_ordinary("SUSPENDED");
  Swaiting = make_ordinary("WAITING");
  Sstopped = make_ordinary("STOPPED");
  Sdead = make_ordinary("DEAD");
  Sblocking = make_ordinary("BLOCKING");
  Sdelayed = make_ordinary("DELAYED");
  siSthread_top_level = make_si_ordinary("THREAD-TOP-LEVEL");

  make_si_function("THREAD-BREAK-IN", siLthread_break_in);
/*  make_si_function("THREAD-BREAK-QUIT", siLthread_break_quit); */
  make_si_function("THREAD-BREAK-RESUME", siLthread_break_resume);

  make_function("MAKE-THREAD", Lmake_thread);
  make_function("DEACTIVATE", Ldeactivate);
  make_function("REACTIVATE", Lreactivate);
  make_function("KILL-THREAD", Lkill_thread);
  
  make_function("CURRENT-THREAD", Lcurrent_thread);
  make_function("THREAD-STATUS", Lthread_status);
  make_function("THREAD-LIST", Lthread_list);
  
  make_function("MAKE-CONTINUATION", Lmake_continuation);
  make_function("THREAD-OF", Lthread_of);
  make_function("CONTINUATION-OF", Lcontinuation_of);
  make_function("RESUME", Lresume);

  make_function("%DISABLE-SCHEDULER", Ldisable_scheduler);
  make_function("%ENABLE-SCHEDULER", Lenable_scheduler);

  make_function("%SUSPEND", Lsuspend);
  make_function("%DELAY", Ldelay);
  make_function("%THREAD-WAIT", Lthread_wait);
  make_function("%THREAD-WAIT-WITH-TIMEOUT", Lthread_wait_with_timeout);
}
