/*
    big.c -- Bignum routines.
*/
/*
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/

#include "config.h"

/* 
 * Using GMP multiple precision integers:
 *
 * typedef struct
 * {
 *  long int alloc;		// Number of limbs allocated and pointed
 *				//   to by the D field.
 *  long int size;		// abs(SIZE) is the number of limbs
 *				//   the last field points to.  If SIZE
 *				//   is negative this is a negative number.
 *  unsigned long int *d;	// Pointer to the limbs,
 *				//   d[0] is the least significative.
 * } MP_INT;
 *
 * typedef unsigned long int	mp_limb;
 *
 */

/* #define alloc_relblock(x, y)	alloc_contblock(x) /* prova */

/*
 * Different from mpz_init since we initialize with NULL limbs
 */
object
big_alloc(int size)
{
  volatile object x = alloc_object(t_bignum);
  x->big.big_dim = size;
  x->big.big_size = 0;
  x->big.big_limbs = (mp_limb *)alloc_relblock(size * sizeof(mp_limb),
					       sizeof(mp_limb));
  return(x);
}

object
bignum1(int val)
{
  volatile object z = alloc_object(t_bignum);
  mpz_init_set_si(&z->big.num, val);
  return(z);
}

object
bignum2(mp_limb hi, mp_limb lo)
{
  object z;

  z = big_alloc(2);
  z->big.big_size = 2;
  z->big.big_limbs[0] = lo;
  z->big.big_limbs[1] = hi;
  return(z);
}

object
big_copy(object x)
{
	volatile object y = big_alloc(abs(x->big.big_size));
	mpz_set(&y->big.num, &x->big.num);
	return(y);
}

/*
	big_zerop(x) tells whether bignum x is zero or not.

#define big_zerop(x)	(mp_size(&x->big.num) == 0)
*/

/*
	big_sign(x) returns
		something < 0	if x < 0
		0		if x = 0
		something > 0	if x > 0.

#define big_sign(x)	(x->big.big_size)
*/

/*
	big_compare(x, y) returns
		-1	if x < y
		0	if x = y
		1	if x > y.

#define big_compare(x, y)	mpz_cmp(&x->big.num, &y->big.num)
*/

/*
	big_complement(x) destructively takes
	the complement of bignum x.

#define big_complement(x)	mpz_neg(&x->big.num, x->big.num);
*/

/*
	big_minus(x) returns the complement of bignum x.
*/
object
big_minus(struct bignum *x)
{
	volatile object y = big_alloc(0);
	mpz_neg(&y->big.num, &x->num);
	return(y);
}

/*
	big_add_ui(x, i) destructively adds non-negative int i
	to bignum x.
	I should be non-negative.

	mpz_add_ui(&x->big.num, &x->big.num, i)
*/

/*
	big_sub_ui(x, i) destructively subtracts non-negative int i
	from bignum x.
	I should be non-negative.

	mpz_sub_ui(&x->big.num, &x->big.num, i)
*/

/*
	big_mul_ui(x, i) destructively multiplies non-negative bignum x
	by non-negative int i.
	I should be non-negative.
	X should be non-negative.

	mpn_mul(&x->big.big_limbs, &x->big.big_limbs, x->big.big_size, &i, 1)
*/

/*
	big_div_ui(x, i) destructively divides non-negative bignum x
	by positive int i.
	X will hold the remainder of the division.
	div_int_big(i, x) returns the remainder of the division.
	I should be positive.
	X should be non-negative.

	mp_limb q[x->big.big_size];
	mpn_div(q, &x->big.big_limbs, &x->big.big_size, &i, 1), x
*/

/*
	big_plus(x, y) returns the sum of bignum x and bignum y.
	X and y may be any bignum.
*/
object
big_plus(struct bignum *x, struct bignum *y)
{
	volatile object z = big_alloc(0);
	mpz_add(&z->big.num, &x->num, &y->num);
	return(z);
}

object
big_normalize(struct bignum *x)
{
  int s = x->big_size;
  mp_limb y;
  if (s == 0)
    return(MAKE_FIXNUM(0));
  y = x->big_limbs[0];
  if (s == 1 && y <= MOST_POSITIVE_FIX)
    return(MAKE_FIXNUM(y));
  if (s == -1 && y <= MOST_POSITIVE_FIX + 1)
    return(MAKE_FIXNUM(-y));
  return((object)x);
}

double
big_to_double(struct bignum *x)
{
	double d, e;
	int i;
	mp_limb *limbs = x->big_limbs;
	int len = abs(x->big_size);

	for (d = 0.0, e = 1.0, i = 0; i < len;  i++) {
		d += e * (double)limbs[i];
		e *= 4.294967296e9;
	}
	if (big_sign((object)x) < 0)
	  d = -d;
	return(d);
}

long
big_to_long(struct bignum *x)
{
	long d = x->big_limbs[0];
	if (big_sign((object)x) < 0)
	  d = -d;
	return(d);
}

void *
mp_alloc(size_t size)
{
  return(alloc_relblock(size, sizeof(mp_limb)));
}

static object realloc_big;

void *
mp_realloc(void *ptr, size_t osize, size_t nsize)
{
  void *p;
  /* GC from alloc_relblock may move ptr */
  realloc_big->big.big_limbs = ptr;
  realloc_big->big.big_dim = osize / sizeof(mp_limb);
  p = alloc_relblock(nsize, sizeof(mp_limb));
  memcpy(p, realloc_big->big.big_limbs, osize);
  realloc_big->big.big_limbs = NULL;
  return(p);
}

void
mp_free(void *ptr, size_t size) {}

init_big()
{
	realloc_big = big_alloc(0);
	register_root(&realloc_big);
	mp_set_memory_functions(mp_alloc, mp_realloc, mp_free);
}
