;;;;  ANSI.LSP  -- Compatibility with ANSI Common Lisp

;;;;  Copyright (c) 1990, Giuseppe Attardi.
;;;;
;;;;    This program is free software; you can redistribute it and/or
;;;;    modify it under the terms of the GNU Library General Public
;;;;    License as published by the Free Software Foundation; either
;;;;    version 2 of the License, or (at your option) any later version.
;;;;
;;;;    See file '../Copyright' for full details.

(in-package 'lisp)
(export '(declaim defpackage print-unreadable-object
	  nth-value complement destructuring-bind with-standard-io-syntax
	  function-lambda-expression))
(in-package 'system)

;;; ----------------------------------------------------------------------
;;; X3J13 vote <108>

(unless (find-package "COMMON-LISP")
  (rename-package 'lisp 'common-lisp '("LISP" "CL")))
(unless (find-package "COMMON-LISP-USER")
  (rename-package 'user 'common-lisp-user '("USER" "CL-USER")))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <144>

(defmacro declaim (&rest decl-specs)
  (if (cdr decl-specs)
      `(progn
	 ,@(mapcar #'(lambda (decl-spec) `(proclaim ',decl-spec))
		   decl-specs))
    `(proclaim ',(car decl-specs))))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <40>

(defmacro print-unreadable-object
	  ((object stream &key type identity) &body body)
  `(progn
    (print-unreadable-object-before ,object ,stream ,type ,(null body))
    ,@body
    (print-unreadable-object-after ,object ,stream ,identity)))

(defun print-unreadable-object-before (object stream type body)
  (princ "#<" stream)
  (when type
    (prin1 (type-of object) stream))
  (unless body
    (princ " " stream)))
  
(defun print-unreadable-object-after (object stream identity)
  (when identity
    (princ " " stream)
    (princ (si:pointer object) stream))
  (princ ">" stream)
  nil)

;;; ----------------------------------------------------------------------
;;; X3J13 vote <52>

;;; Package definition and installation, CLtL2 S. 270
;;; from Bruno Haible
(defmacro defpackage (packname &rest options)
  (flet ((stringify (x)
	   (etypecase x
	     (string x)
	     (symbol (symbol-name x)))))
    (setq packname (stringify packname))
					; Options:
    (let ((size nil)			; Flag :SIZE not handled
          (documentation nil)		; Flag :DOCUMENTATION not handled
          (nickname-list '())		; List of Nicknames
          (shadow-list '())		; List of Symbols
          (shadowing-list '())		; List of pairs (Symbol . Package) for shadowing-import
          (use-list '())		; List of Packages for use-package
          (use-default '("LISP"))	; Default for use-list
          (import-list '())		; List of pairs (Symbol . Package) for import
          (intern-list '())		; List of Symbols for intern
          (symname-list '())		; List all Symbols mentioned
          (export-list '()))		; List of Symbols for export
      (flet ((record-symname (name)
               (if (member name symname-list :test #'string=)
		   (error
		    "~S ~A: the symbol ~A must not be specified more than once"
		    'defpackage packname name)
		   (push name symname-list))
               name))
        (dolist (option options)
          (if (listp option)
	      (if (keywordp (car option))
		  (case (first option)
		    (:SIZE
		     (if size
			 (error
			  "~S ~A: the ~S option must not be given more than once"
			  'defpackage packname ':SIZE)
			 (setq size t)))	; ignore further argument
		    (:DOCUMENTATION	; dpANS
		     (if documentation
			 (error
			  "~S ~A: the ~S option must not be given more than once"
			  'defpackage packname ':DOCUMENTATION)
			 (setq documentation t))) ; ignore further argument
		    (:NICKNAMES
		     (dolist (name (rest option))
		       (push (stringify name) nickname-list)))
		    (:SHADOW
		     (dolist (name (rest option))
		       (push (record-symname (stringify name)) shadow-list)))
		    (:SHADOWING-IMPORT-FROM
		     (let ((pack (stringify (second option))))
		       (dolist (name (cddr option))
			 (push (cons (record-symname (stringify name)) pack)
			       shadowing-list))))
		    (:USE
		     (dolist (name (rest option))
		       (push (stringify name) use-list))
		     (setq use-default nil))
		    (:IMPORT-FROM
		     (let ((pack (stringify (second option))))
		       (dolist (name (cddr option))
			 (push (cons (record-symname (stringify name)) pack)
			       import-list))))
		    (:INTERN
		     (dolist (name (rest option))
		       (push (record-symname (stringify name)) intern-list)))
		    (:EXPORT
		     (dolist (name (rest option))
		       (push (stringify name) export-list)))
		    (T (error
			"~S ~A: unknown option ~S"
			'defpackage packname (first option))))
		  (error
		   "~S ~A: invalid syntax in ~S option: ~S"
		   'defpackage packname 'defpackage option))
	      (error
	       "~S ~A: not a ~S option: ~S"
	       'defpackage packname 'defpackage option)))
	;; check no intersection between intern-list and export-list:
        (setq symname-list intern-list)
        (mapc #'record-symname export-list))
      ;; Reverse the lists and compute use-list:
      (setq nickname-list (nreverse nickname-list))
      (setq shadow-list (nreverse shadow-list))
      (setq shadowing-list (nreverse shadowing-list))
      (setq use-list (or use-default (nreverse use-list)))
      (setq import-list (nreverse import-list))
      (setq intern-list (nreverse intern-list))
      (setq export-list (nreverse export-list))
      ;; Expansion:
      `(EVAL-WHEN (LOAD COMPILE EVAL)
	;; First make sure the package exists
	(%make-package ,packname ',nickname-list)
	;; Shadow stuff next
	,@(if shadow-list
	      `((SHADOW ',(mapcar #'make-symbol shadow-list) ,packname)))
	,@(mapcar
	   #'(lambda (pair)
	       `(SHADOWING-IMPORT-CERROR ,(car pair) ,(cdr pair) ,packname))
	   shadowing-list)
	;; Step 2
	,@(if use-list `((USE-PACKAGE ',use-list ,packname)))
	;; Step 3
	,@(mapcar
	   #'(lambda (pair)
	       `(IMPORT-CERROR ,(car pair) ,(cdr pair) ,packname))
	   import-list)
	,@(mapcar
	   #'(lambda (symname) `(INTERN ,symname ,packname))
	   intern-list)
	;; Step 4
	,@(if export-list
	      `((INTERN-EXPORT ',export-list ,packname)))
	(FIND-PACKAGE ,packname)))))

;;; Auxiliary functions:

(defun %make-package (packname nickname-list)
  (let ((p (find-package packname)))
    (if p
	(rename-package p packname nickname-list)
	(make-package packname :NICKNAMES nickname-list :USE '()))))

(defun find-symbol-cerror (string packname calling-packname)
  (multiple-value-bind (sym found) (find-symbol string packname)
    (unless found
      (cerror "This symbol will be created."
	      "~S ~A: There is no symbol ~A::~A ."
              'defpackage calling-packname packname string)
      (setq sym (intern string packname)))
    sym))

(defun shadowing-import-cerror (string packname calling-packname)
  (shadowing-import (find-symbol-cerror string packname calling-packname)
                    calling-packname))
(defun import-cerror (string packname calling-packname)
  (import (find-symbol-cerror string packname calling-packname)
          calling-packname))
(defun intern-export (string-list packname)
  (export (mapcar #'(lambda (string) (intern string packname)) string-list)
          packname))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <123>

;; Macro (nth-value n form) == (nth n (multiple-value-list form)), CLtL2 S. 184
(defmacro nth-value (n form)
  (if (and (integerp n) (>= n 0))
    (if (< n (1- multiple-values-limit))
      (if (= n 0)
        `(PROG1 ,form)
        (let ((resultvar (gensym)))
          (do ((vars (list resultvar))
               (ignores nil)
               (i n (1- i)))
              ((zerop i)
               `(MULTIPLE-VALUE-BIND ,vars ,form
                  (DECLARE (IGNORE ,@ignores))
                  ,resultvar))
            (let ((g (gensym))) (push g vars) (push g ignores)))))
      `(PROGN ,form NIL))
    `(NTH ,n (MULTIPLE-VALUE-LIST ,form))))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <64>

;;; Exploit defmacro expansion.
;;; Unfortunately error messages mention defmacro-lambda-list.
(defmacro destructuring-bind (pattern form &body body)
  (multiple-value-bind (lambda doc pprint)
      (sys:expand-defmacro nil pattern body)
    (declare (ignore doc pprint))
    (setq lambda (third lambda))
    `(let* ((,(car lambda) (cons nil ,form))
	    ,@(cdddr lambda))
      ,@body)))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <87>

(defun complement (fun)
  #'(lambda (&rest arguments) (not (apply fun arguments))))

;;; ----------------------------------------------------------------------
;;; part of X3J13 vote <98>

(defconstant *common-lisp-user-package* (find-package "COMMON-LISP-USER"))

(defmacro with-standard-io-syntax (&body body &environment env)
  (multiple-value-bind (declarations body-rest)
      (sys::find-declarations body)
    ;; It would be possible to put all these bindings into a single function,
    ;; but this would force variables into closures.
    `(LET (; printer/reader variables:
           (*PACKAGE*                   *COMMON-LISP-USER-PACKAGE*)
           ; printer variables:
           (*PRINT-ARRAY*               T)
           (*PRINT-BASE*                10)
           (*PRINT-CASE*                ':UPCASE)
           (*PRINT-CIRCLE*              NIL)
           (*PRINT-ESCAPE*              T)
           (*PRINT-GENSYM*              T)
           (*PRINT-LENGTH*              NIL)
           (*PRINT-LEVEL*               NIL)
          ;(*PRINT-LINES*               NIL) ; XP variable not present in ECL
          ;(*PRINT-MISER-WIDTH*         NIL) ; XP variable not present in ECL
          ;(*PRINT-PPRINT-DISPATCH*     NIL) ; XP variable not present in ECL
           (*PRINT-PRETTY*              NIL)
           (*PRINT-RADIX*               NIL)
           (*PRINT-READABLY*            T)
          ;(*PRINT-RIGHT-MARGIN*        NIL) ; XP variable not present in ECL
           ; reader variables:
           (*READ-BASE*                 10)
           (*READ-DEFAULT-FLOAT-FORMAT* 'SINGLE-FLOAT)
           (*READ-EVAL*                 T)
           (*READ-SUPPRESS*             NIL)
           (*READTABLE*                 (COPY-READTABLE NIL)))
       ,@(if declarations `((DECLARE ,@declarations)))
       ,@body-rest)))

;;; ----------------------------------------------------------------------
;;; X3J13 vote <88>

;; Returns originating expression for a function, CLtL2 S. 682
(defun function-lambda-expression (obj)
  (if (functionp obj)
      (if (compiled-function-p obj)	; compiled function?
	  (values nil t nil)
	  (let (name def env)
	    (case (car obj)
	      (LAMBDA (setq def (cdr obj)))
	      (LAMBDA-BLOCK (setq name (second def)
				  def (cddr obj)))
	      (LAMBDA-CLOSURE
	       (setq env (second obj)
		     def (cddddr obj)))
	      (LAMBDA-BLOCK-CLOSURE
	       (setq env (second obj)
		     def (cddddr obj)
		     name (first def)
		     def (rest def))))
	    (values (cons 'LAMBDA def) env name)))
      (error "~S: ~S is not a function" 'function-lambda-expression obj)))
