;;;;  Copyright (c) 1992, Giuseppe Attardi.
;;;;
;;;;    This program is free software; you can redistribute it and/or
;;;;    modify it under the terms of the GNU Library General Public
;;;;    License as published by the Free Software Foundation; either
;;;;    version 2 of the License, or (at your option) any later version.
;;;;
;;;;    See file '../Copyright' for full details.
;;;
;;; conditions.lsp
;;;
;;; Originally written by Kent M. Pitman of Symbolics, Inc. and
;;; distributed without any copyright.
;;; This is Version 18.
;;;
;;; KMP's disclaimer:
;;;
;;; This is a sample implementation. It is not in any way intended as the
;;; definition of any aspect of the condition system. It is simply an existence
;;; proof that the condition system can be implemented.
;;;

#-CLOS
(eval-when (compile load eval)
  (print
   "Without CLOS this does not implement the slot syntax required by CLtL2"))

(in-package "CONDITIONS" :USE '("LISP"))


(import 'SI::ABORT)			; defined in "SYSTEM" by setdoc.lsp

(export '(*break-on-signals* *debugger-hook* signal
	  handler-case handler-bind ignore-errors define-condition make-condition
	  with-simple-restart restart-case restart-bind restart-name
	  restart-name find-restart compute-restarts invoke-restart
	  invoke-restart-interactively abort continue muffle-warning
	  store-value use-value invoke-debugger restart condition
	  warning serious-condition simple-condition simple-warning simple-error
	  simple-condition-format-string simple-condition-format-arguments
	  storage-condition stack-overflow storage-exhausted type-error
	  type-error-datum type-error-expected-type simple-type-error
	  program-error control-error stream-error stream-error-stream
	  end-of-file file-error file-error-pathname cell-error
	  unbound-variable undefined-function arithmetic-error
	  arithmetic-error-operation arithmetic-error-operands
	  package-error package-error-package
	  division-by-zero floating-point-overflow floating-point-underflow))

;;; ----------------------------------------------------------------------
;;; Unique Ids

(defmacro unique-id (obj)
  "Generates a unique integer ID for its argument."
  `(sys:pointer ,obj))


;;; Restarts

(defvar *restart-clusters* '())

;;; do we need copy-list if *restart-clusters* has only one element? Beppe
(defun compute-restarts () (apply #'append *restart-clusters*))

(defun restart-print (restart stream depth)
  (declare (ignore depth))
  (if *print-escape*
      (format stream "#<~s.~d>" (type-of restart) (unique-id restart))
      (restart-report restart stream)))

(defstruct (restart (:PRINT-FUNCTION restart-print))
  name
  function
  report-function
  interactive-function)

(defun restart-report (restart stream)
  (let ((fn (restart-report-function restart)))
    (if fn
	(funcall fn stream)
	(format stream "~s" (or (restart-name restart) restart)))))

(defmacro restart-bind (bindings &body forms)
  `(let ((*restart-clusters*
	  (cons (list ,@(mapcar #'(lambda (binding)
				    `(make-restart
				      :NAME     ',(car binding)
				      :FUNCTION ,(cadr binding)
				      ,@(cddr binding)))
				bindings))
		*restart-clusters*)))
     ,@forms))

(defun find-restart (name)
  (dolist (restart-cluster *restart-clusters*)
    (dolist (restart restart-cluster)
      (when (or (eq restart name) (eq (restart-name restart) name))
	(return-from find-restart restart)))))
  
(defun invoke-restart (restart &rest values)
  (let ((real-restart (or (find-restart restart)
			  (error "Restart ~S is not active." restart))))
    (apply (restart-function real-restart) values)))

(defun invoke-restart-interactively (restart)
  (let ((real-restart (or (find-restart restart)
			  (error "Restart ~S is not active." restart))))
    (apply (restart-function real-restart)
	   (let ((interactive-function
		   (restart-interactive-function real-restart)))
	     (if interactive-function
		 (funcall interactive-function)
		 '())))))


(defmacro restart-case (expression &body clauses)
  (flet ((transform-keywords (&key report interactive)
	   (let ((keywords '()))
	     (when interactive
	       (setq keywords (list :INTERACTIVE-FUNCTION
				    `#',interactive)))
	     (when report
	       (setq keywords (list* :REPORT-FUNCTION
				     (if (stringp report)
					 `#'(lambda (stream)
					      (write-string ,report stream))
					 `#',report)
				     keywords)))
	     keywords)))
    (let ((block-tag (gensym))
	  (temp-var  (gensym))
	  (data (mapcar #'(lambda (clause)
			    (let (keywords (forms (cddr clause)))
			      (do ()
				  ((null forms))
				(if (keywordp (car forms))
				    (setq keywords (list* (car forms)
							  (cadr forms)
							  keywords)
					  forms (cddr forms))
				    (return)))
			      (list (car clause) 		;Name=0
				    (gensym) 			;Tag=1
				    (apply #'transform-keywords ;Keywords=2
					   keywords)
				    (cadr clause)		;BVL=3
				    forms))) 			;Body=4
			clauses)))
      `(block ,block-tag
	 (let ((,temp-var nil))
	   (tagbody
	     (restart-bind
	       ,(mapcar #'(lambda (datum)
			    (let ((name (nth 0 datum))
				  (tag  (nth 1 datum))
				  (keys (nth 2 datum)))
			      `(,name #'(lambda (&rest temp)
					  (setq ,temp-var temp)
					  (go ,tag))
				,@keys)))
			data)
	       (return-from ,block-tag ,expression))
	     ,@(mapcan #'(lambda (datum)
			   (let ((tag  (nth 1 datum))
				 (bvl  (nth 3 datum))
				 (body (nth 4 datum)))
			     (list tag
				   `(return-from ,block-tag
				      (apply #'(lambda ,bvl ,@body)
					     ,temp-var)))))
		       data)))))))

(defmacro with-simple-restart ((restart-name format-string
					     &rest format-arguments)
			       &body forms)
  `(restart-case (progn ,@forms)
     (,restart-name ()
        :REPORT (lambda (stream)
		  (format stream ,format-string ,@format-arguments))
      (values nil t))))


;;; ----------------------------------------------------------------------
;;; Condition Data Type

(defun condition-print (condition stream depth)
  (declare (ignore depth))
  (if *print-escape*
      (format stream "#<~S.~D>" (type-of condition) (unique-id condition))
      (condition-report condition stream)))

(defstruct (condition :CONC-NAME
                      (:CONSTRUCTOR |constructor for condition|)
                      (:PREDICATE nil)
                      (:PRINT-FUNCTION condition-print)))

#-CLOS
(progn
(eval-when (eval compile load)

(defmacro parent-type     (condition-type) `(get ,condition-type 'PARENT-TYPE))
(defmacro slots           (condition-type) `(get ,condition-type 'SLOTS))
(defmacro conc-name       (condition-type) `(get ,condition-type 'CONC-NAME))
(defmacro report-function (condition-type) `(get ,condition-type 'REPORT-FUNCTION))
(defmacro make-function   (condition-type) `(get ,condition-type 'MAKE-FUNCTION))

);nehw-lave

(defun condition-report (condition stream)
  (do ((type (type-of condition) (parent-type type)))
      ((not type) (format stream "The condition ~A occurred."))
    (let ((reporter (report-function type)))
      (when reporter
        (funcall reporter condition stream)
        (return nil)))))

(setf (make-function   'CONDITION) '|constructor for condition|)

(defun make-condition (type &rest slot-initializations)
  (let ((fn (make-function type)))
    (if fn
	(apply fn slot-initializations)
	(error 'SIMPLE-TYPE-ERROR
	       :DATUM type
	       :EXPECTED-TYPE '(SATISFIES MAKE-FUNCTION)
	       :FORMAT-STRING "Not a condition type: ~S"
	       :FORMAT-ARGUMENTS (list type)))))

(eval-when (eval compile load) ;Some utilities that are used at macro expansion time

(defmacro resolve-function (function expression resolver)
  `(cond ((and ,function ,expression)
          (cerror "Use only the :~A information."
                  "Only one of :~A and :~A is allowed."
                  ',function ',expression))
         (,expression
          (setq ,function ,resolver))))
         
(defun parse-new-and-used-slots (slots parent-type)
  (let ((new '()) (used '()))
    (dolist (slot slots)
      (if (slot-used-p (car slot) parent-type)
          (push slot used)
          (push slot new)))
    (values new used)))

(defun slot-used-p (slot-name type)
  (cond ((eq type 'CONDITION) nil)
        ((not type)
	 (error "The type ~S does not inherit from CONDITION." type))
        ((assoc slot-name (slots type)))
        (t
         (slot-used-p slot-name (parent-type type)))))

);nehw-lave

(defmacro define-condition (name (parent-type) slot-specs &rest options)
  (let ((constructor
	 (let ((*package* (find-package "CONDITIONS"))) ;Bind for the INTERN -and- the FORMAT
	   (intern (format nil "Constructor for ~S" name)))))
    (let ((slots (mapcar #'(lambda (slot-spec)
			     (if (atom slot-spec) (list slot-spec) slot-spec))
			 slot-specs)))
      (multiple-value-bind (new-slots used-slots)
          (parse-new-and-used-slots slots parent-type)
	(let ((conc-name-p     nil)
	      (conc-name       nil)
	      (report-function nil)
	      (documentation   nil))
	  (dolist (option options)
	    (case (car option)		; should be ecase
	      (:CONC-NAME (setq conc-name-p t)
			  (setq conc-name (cadr option)))
	      (:REPORT (setq report-function
			     (if (stringp (cadr option))
				 `(lambda (stream)
				   (write-string ,(cadr option) stream))
				 (cadr option))))
	      (:DOCUMENTATION (setq documentation (cadr option)))
	      (otherwise (cerror "Ignore this DEFINE-CONDITION option."
				 "Invalid DEFINE-CONDITION option: ~S" option))))
	  (unless conc-name-p
	    (setq conc-name (intern (format nil "~A-" name) *PACKAGE*)))
          ;; The following three forms are compile-time side-effects. For now, they affect
          ;; the global environment, but with modified abstractions for PARENT-TYPE, SLOTS, 
          ;; and CONC-NAME, the compiler could easily make them local.
          (setf (parent-type name) parent-type)
          (setf (slots name)       slots)
          (setf (conc-name name)   conc-name)
          ;; Finally, the expansion ...
          `(PROGN (DEFSTRUCT (,NAME
			       (:CONSTRUCTOR ,CONSTRUCTOR)
			       (:PREDICATE NIL)
			       (:COPIER NIL)
			       (:PRINT-FUNCTION CONDITION-PRINT)
			       (:INCLUDE ,PARENT-TYPE ,@USED-SLOTS)
			       (:CONC-NAME ,CONC-NAME))
                    ,@NEW-SLOTS)
	    (SETF (DOCUMENTATION ',NAME 'TYPE) ',DOCUMENTATION)
	    (SETF (PARENT-TYPE ',NAME) ',PARENT-TYPE)
	    (SETF (SLOTS ',NAME) ',SLOTS)
	    (SETF (CONC-NAME ',NAME) ',CONC-NAME)
	    (SETF (REPORT-FUNCTION ',NAME) ,(IF REPORT-FUNCTION `#',REPORT-FUNCTION))
	    (SETF (MAKE-FUNCTION ',NAME) ',CONSTRUCTOR)
	    ',NAME))))))
)

#+CLOS
(progn
(defmacro define-condition (name parent-list slot-specs &rest options)
  (let* ((report-function nil)
	 (documentation nil))
    (dolist (option options)
      (case (car option)
	(:REPORT (setq report-function
		       (when (stringp (cadr option))
			 `(LAMBDA (CONDITION STREAM)
			   (DECLARE (IGNORE CONDITION))
			   (WRITE-STRING ,(CADR OPTION) STREAM))
			 (CADR OPTION))))
	(:DOCUMENTATION (setq documentation (cadr option)))
	(otherwise (cerror "Ignore this DEFINE-CONDITION option."
			   "Invalid DEFINE-CONDITION option: ~S" option))))
    `(PROGN
      (EVAL-WHEN (COMPILE))
      (DEFCLASS ,name ,parent-list ,slot-specs)
      (EVAL-WHEN (COMPILE LOAD EVAL)
	(SETF (GET ',name 'DOCUMENTATION) ',documentation))
      ,@(when report-function
	      `((DEFMETHOD PRINT-OBJECT ((X ,name) STREAM)
		    (IF *PRINT-ESCAPE*
			(CALL-NEXT-METHOD)
			(,report-function X STREAM)))))
      ',NAME)))


(defun make-condition (type &rest slot-initializations)
  (unless (subtypep type 'CONDITION)
    (error 'SIMPLE-TYPE-ERROR
	   :DATUM type
	   :EXPECTED-TYPE 'CONDITION
	   :FORMAT-STRING "Not a condition type: ~S"
	   :FORMAT-ARGUMENTS (list type)))
  (apply #'make-instance type slot-initializations))

)

#| For the moment, do not redefine these. Beppe.
(eval-when (eval compile load)

(defun accumulate-cases (macro-name cases list-is-atom-p)
  (do ((c cases (cdr c))
       (l '()))
      ((null c) (nreverse l))
    (let ((keys (caar c)))
      (cond ((atom keys)
	     (cond ((null keys))
		   ((member keys '(OTHERWISE T))
		    (error "OTHERWISE is not allowed in ~S expressions."
			   macro-name))
		   (t (push keys l))))
	    (list-is-atom-p
	     (push keys l))
	    (t (setq l (append keys l)))))))

);nehw-lave

(defmacro ecase (keyform &rest cases)
  (let ((keys (accumulate-cases 'ECASE cases nil)))
    `(case ,keyform
      ,@cases
      (otherwise
       (error 'CASE-FAILURE :name 'ECASE
	:datum ,keyform
	:expected-type '(MEMBER ,@keys)
	:possibilities ',keys))))))

(defmacro ccase (keyplace &rest cases)
  (let ((keys (accumulate-cases 'CCASE cases nil))
	(tag1 (gensym))
	(tag2 (gensym)))
    `(block ,tag1
       (tagbody ,tag2
	 (return-from ,tag1
	   (case ,keyplace
	     ,@cases
	     (otherwise
	       (restart-case (error 'CASE-FAILURE
				    :name 'CCASE
				    :datum ,keyplace
				    :expected-type '(MEMBER ,@keys)
				    :possibilities ',keys)
		 (store-value (value)
		     :report (lambda (stream)
			       (format stream "Supply a new value of ~S."
				       ',keyplace))
		     :interactive read-evaluated-form
		   (setf ,keyplace value)
		   (go ,tag2))))))))))



(defmacro etypecase (keyform &rest cases)
  (let ((types (accumulate-cases 'ETYPECASE cases t)))
    `(typecase ,keyform
      ,@cases
      (otherwise
       (error 'CASE-FAILURE :name 'ETYPECASE
	:datum ,keyform
	:expected-type '(OR ,@types)
	:possibilities ',types)))))

(defmacro ctypecase (keyplace &rest cases)
  (let ((types (accumulate-cases 'CTYPECASE cases t))
	(tag1 (gensym))
	(tag2 (gensym)))
    `(block ,tag1
       (tagbody ,tag2
	 (return-from ,tag1
	   (typecase ,keyplace
	     ,@cases
	     (otherwise
	       (restart-case (error 'CASE-FAILURE
				    :name 'CTYPECASE
				    :datum ,keyplace
				    :expected-type '(OR ,@types)
				    :possibilities ',types)
		 (store-value (value)
		     :REPORT (lambda (stream)
			       (format stream "Supply a new value of ~S."
				       ',keyplace))
		     :INTERACTIVE read-evaluated-form
		   (setf ,keyplace value)
		   (go ,tag2))))))))))

|#


(defun assert-report (names stream)
  (format stream "Retry assertion")
  (if names
      (format stream " with new value~P for ~{~S~^, ~}."
	      (length names) names)
      (format stream ".")))

(defun assert-prompt (name value)
  (cond ((y-or-n-p "The old value of ~S is ~S.~
		  ~%Do you want to supply a new value? "
		   name value)
	 (format *query-io* "~&type a form to be evaluated:~%")
	 (flet ((read-it () (eval (read *query-io*))))
	   (if (symbolp name) ;Help user debug lexical variables
	       (progv (list name) (list value) (read-it))
	       (read-it))))
	(t value)))

(defun simple-assertion-failure (assertion)
  (error 'SIMPLE-TYPE-ERROR
	 :DATUM assertion
	 :EXPECTED-TYPE nil		; This needs some work in next revision. -kmp
	 :FORMAT-STRING "The assertion ~S failed."
	 :FORMAT-ARGUMENTS (list assertion)))

(defmacro assert (test-form &optional places datum &rest arguments)
  (let ((tag (gensym)))
    `(tagbody ,tag
       (unless ,test-form
	 (restart-case ,(if datum
			    `(error ,datum ,@arguments)
			    `(simple-assertion-failure ',test-form))
	   (continue ()
	       :REPORT (lambda (stream) (assert-report ',places stream))
	     ,@(mapcar #'(lambda (place)
			   `(setf ,place (assert-prompt ',place ,place)))
		       places)
             (go ,tag)))))))



(defun read-evaluated-form ()
  (format *query-io* "~&Type a form to be evaluated:~%")
  (list (eval (read *query-io*))))

(defmacro check-type (place type &optional type-string)
  (let ((tag1 (gensym))
	(tag2 (gensym)))
    `(block ,tag1
       (tagbody ,tag2
	 (if (typep ,place ',type) (return-from ,tag1 nil))
	 (restart-case ,(if type-string
			    `(error "The value of ~S is ~S, ~
				     which is not ~A."
				    ',place ,place ,type-string)
			    `(error "The value of ~S is ~S, ~
				     which is not of type ~S."
				    ',place ,place ',type))
	   (store-value (value)
	       :REPORT (lambda (stream)
			 (format stream "Supply a new value of ~S."
				 ',place))
	       :INTERACTIVE read-evaluated-form
	     (setf ,place value)
	     (go ,tag2)))))))

(defvar *handler-clusters* nil)

(defmacro handler-bind (bindings &body forms)
  (unless (every #'(lambda (x) (and (listp x) (= (length x) 2))) bindings)
    (error "Ill-formed handler bindings."))
  `(let ((*handler-clusters*
	  (cons (list ,@(mapcar #'(lambda (x) `(cons ',(car x) ,(cadr x)))
				bindings))
		*handler-clusters*)))
     ,@forms))

(defvar *break-on-signals* nil)

(defun signal (datum &rest arguments)
  (let ((condition
	 (coerce-to-condition datum arguments 'SIMPLE-CONDITION 'SIGNAL))
        (*handler-clusters* *handler-clusters*))
    (if (typep condition *break-on-signals*)
	(break "~A~%Break entered because of *BREAK-ON-SIGNALS*."
	       condition))
    (loop (unless *handler-clusters* (return))
          (let ((cluster (pop *handler-clusters*)))
	    (dolist (handler cluster)
	      (when (typep condition (car handler))
		(funcall (cdr handler) condition)
		(return nil) ;?
		))))
    nil))



;;; COERCE-TO-CONDITION
;;;  Internal routine used in ERROR, CERROR, BREAK, and WARN for parsing the
;;;  hairy argument conventions into a single argument that's directly usable 
;;;  by all the other routines.

(defun coerce-to-condition (datum arguments default-type function-name)
  (cond ((typep datum 'CONDITION)
	 (when arguments
	   (cerror "Ignore the additional arguments."
		   'SIMPLE-TYPE-ERROR
		   :DATUM arguments
		   :EXPECTED-TYPE 'NULL
		   :FORMAT-STRING "You may not supply additional arguments ~
				     when giving ~S to ~S."
		   :FORMAT-ARGUMENTS (list datum function-name)))
	 datum)
        ((symbolp datum)                  ;roughly, (subtypep datum 'CONDITION)
         (apply #'make-condition datum arguments))
        ((stringp datum)
	 (make-condition default-type
                         :FORMAT-STRING datum
                         :FORMAT-ARGUMENTS arguments))
        (t
         (error 'SIMPLE-TYPE-ERROR
		:DATUM datum
		:EXPECTED-TYPE '(OR SYMBOL STRING)
		:FORMAT-STRING "Bad argument to ~S: ~S"
		:FORMAT-ARGUMENTS (list function-name datum)))))

(defun error (datum &rest arguments)
  (let ((condition (coerce-to-condition datum arguments 'SIMPLE-ERROR 'ERROR)))
    (signal condition)
    (invoke-debugger condition)))

(defun cerror (continue-string datum &rest arguments)
  (with-simple-restart
	(continue "~A" (format nil "~?" continue-string arguments))
      (apply #'error datum arguments))
  nil)

(defun break (&optional (format-string "Break") &rest format-arguments)
  (with-simple-restart (continue "Return from BREAK.")
    (invoke-debugger
      (make-condition 'SIMPLE-CONDITION
		      :FORMAT-STRING    format-string
		      :FORMAT-ARGUMENTS format-arguments)))
  nil)

(define-condition warning (condition) ())

(defun warn (datum &rest arguments)
  (let ((condition
	  (coerce-to-condition datum arguments 'SIMPLE-WARNING 'WARN)))
    (check-type condition warning "a warning condition")
    (if *break-on-warnings*
	(break "~A~%Break entered because of *BREAK-ON-WARNINGS*."
	       condition))
    (restart-case (signal condition)
      (muffle-warning ()
	  :REPORT "Skip warning."
	(return-from warn nil)))
    (format *error-output* "~&Warning:~%~A~%" condition)
    nil))



(define-condition serious-condition (condition) ())

(define-condition error (serious-condition) ())

(defun simple-condition-printer (condition stream)
  (format stream "~?" (simple-condition-format-string    condition)
	  (simple-condition-format-arguments condition)))

(define-condition simple-condition (condition)
  #-CLOS (format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%simple-condition-)
  #+CLOS ((format-string :INITARG :FORMAT-STRING :INITFORM ""
			 :ACCESSOR format-string)
	  (format-arguments :INITARG :FORMAT-ARGUMENTS
			    :ACCESSOR format-arguments))
  (:REPORT simple-condition-printer))

(define-condition simple-warning (#+CLOS simple-condition warning)
  #-CLOS (format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%simple-warning-)
  #-CLOS (:REPORT simple-condition-printer)
  #+CLOS ())

(define-condition simple-error (#+CLOS simple-condition error)
  #-CLOS (format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%simple-error-)
  #-CLOS (:REPORT simple-condition-printer)
  #+CLOS ())

(define-condition storage-condition (serious-condition) ())

(define-condition stack-overflow    (storage-condition) ())

(define-condition storage-exhausted (storage-condition) ())

(define-condition type-error (error)
  #-CLOS (datum expected-type)
  #+CLOS ((datum :INITARG :DATUM
		 :READER type-error-datum)
	  (expected-type :INITARG :EXPECTED-TYPE
			 :READER type-error-expected-type))
  (:REPORT
   (lambda (condition stream)
     (format stream "~S is not of type ~S."
	     (type-error-datum condition)
	     (type-error-expected-type condition)))))

(define-condition simple-type-error (#+CLOS simple-condition type-error)
  #-CLOS (format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%simple-type-error-)
  #-CLOS (:REPORT simple-condition-printer)
  #+CLOS ())

(define-condition case-failure (type-error)
  #-CLOS (name possibilities)
  #+CLOS ((name :INITARG :NAME
		:READER case-failure-name)
	  (possibilities :INITARG :POSSIBILITIES
			 :READER case-failure-possibilities))
  (:REPORT
   (lambda (condition stream)
     (format stream "~S fell through ~S expression.~%Wanted one of ~:S."
	     (type-error-datum condition)
	     (case-failure-name condition)
	     (case-failure-possibilities condition)))))

(define-condition program-error (error) ())

(define-condition control-error (error) ())

(define-condition stream-error (error)
  #-CLOS (stream)
  #+CLOS ((stream :INITARG :STREAM :READER stream-error-stream)))

(define-condition end-of-file (stream-error)
  ()
  (:REPORT (lambda (condition stream)
	     (format stream "Unexpected end of file on ~S."
		     (stream-error-stream condition)))))

(define-condition file-error (error)
  #-CLOS (pathname)
  #+CLOS ((pathname :INITARG :PATHNAME
		    :READER file-error-pathname)))

(define-condition package-error (error)
  #-CLOS (package)
  #+CLOS ((package :INITARG :PACKAGE
		   :READER package-error-package)))

(define-condition cell-error (error)
  #-CLOS (name)
  #+CLOS ((name :INITARG :NAME
		:READER cell-error-name)))

(define-condition unbound-variable (cell-error)
  ()
  (:REPORT (lambda (condition stream)
	     (format stream "The variable ~S is unbound."
		     (cell-error-name condition)))))
  
(define-condition undefined-function (cell-error)
  ()
  (:REPORT (lambda (condition stream)
	     (format stream "The function ~S is undefined."
		     (cell-error-name condition)))))

(define-condition arithmetic-error (error)
  #-CLOS (operation operands)
  #+CLOS ((operation :INITARG :OPERATION
		     :READER arithmetic-error-operation)))

(define-condition division-by-zero         (arithmetic-error) ())

(define-condition floating-point-overflow  (arithmetic-error) ())

(define-condition floating-point-underflow (arithmetic-error) ())

(define-condition abort-failure (control-error) ()
  (:REPORT (lambda (c s) (declare (ignore c))
		   (write-string "Abort failed." s))))
#+ECL
(progn
(define-condition internal-error (error)
  #-CLOS ((function-name nil))
  #+CLOS ((function-name :INITARG :FUNCTION-NAME
			 :READER internal-error-function-name
			 :INITFORM 'nil))
  (:REPORT (lambda (condition stream)
	     (when (internal-error-function-name condition)
	       (format stream "Error in ~S [or a callee]: "
		       (internal-error-function-name condition)))
	     #+CLOS (call-next-method))))

(defun internal-simple-error-printer (condition stream)
  (when (internal-error-function-name condition)
    (format stream "Error in ~S [or a callee]: "
	    (internal-error-function-name condition)))
  (apply #'format stream (simple-condition-format-string    condition)
	 		 (simple-condition-format-arguments condition)))

(define-condition internal-simple-error
    (internal-error #+CLOS simple-condition)
  #-CLOS ((function-name nil) format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%internal-simple-error-)
  #+CLOS ()
  (:REPORT internal-simple-error-printer))

(define-condition internal-type-error 
    (#+CLOS internal-error type-error)
  #-CLOS ((function-name nil))
  #-CLOS (:CONC-NAME %%internal-type-error-)
  #-CLOS (:REPORT (lambda (condition stream)
		    (when (internal-error-function-name condition)
		      (format stream "Error in ~S [or a callee]: "
			      (internal-error-function-name condition)))
		    (format stream "~S is not of type ~S."
			    (type-error-datum condition)
			    (type-error-expected-type condition))))
  #+CLOS ())

(define-condition internal-simple-program-error 
    (#+CLOS internal-simple-error program-error)
  #-CLOS ((function-name nil) format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%internal-simple-program-error-)
  #-CLOS (:REPORT internal-simple-error-printer)
  #+CLOS ())

(define-condition internal-simple-control-error 
    (#+CLOS internal-simple-error control-error)
  #-CLOS ((function-name nil) format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%internal-simple-control-error-)
  #-CLOS (:REPORT internal-simple-error-printer)
  #+CLOS ())

(define-condition internal-unbound-variable 
    (#+CLOS internal-error unbound-variable)
  #-CLOS ((function-name nil))
  #-CLOS (:CONC-NAME %%internal-unbound-variable-)
  #-CLOS (:REPORT (LAMBDA (CONDITION STREAM)
		    (when (internal-error-function-name condition)
		      (format stream "Error in ~S [or a callee]: "
			      (internal-error-function-name condition)))
		    (FORMAT STREAM "The variable ~S is unbound."
			    (CELL-ERROR-NAME CONDITION))))
  #+CLOS ())

(define-condition internal-undefined-function 
    (#+CLOS internal-error undefined-function)
  #-CLOS ((function-name nil))
  #-CLOS (:CONC-NAME %%internal-undefined-function-)
  #-CLOS (:REPORT (LAMBDA (CONDITION STREAM)
		    (when (internal-error-function-name condition)
		      (format stream "Error in ~S [or a callee]: "
			      (internal-error-function-name condition)))
		    (FORMAT STREAM "The function ~S is undefined."
			    (CELL-ERROR-NAME CONDITION))))
  #+CLOS ())

(define-condition internal-end-of-file 
    (#+CLOS internal-error end-of-file)
  #-CLOS ((function-name nil))
  #-CLOS (:CONC-NAME %%internal-end-of-file-)
  #-CLOS (:REPORT (LAMBDA (CONDITION STREAM)
		    (when (internal-error-function-name condition)
		      (format stream "Error in ~S [or a callee]: "
			      (internal-error-function-name condition)))
		    (FORMAT STREAM "Unexpected end of file on ~S."
			    (STREAM-ERROR-STREAM CONDITION))))
  #+CLOS ())

(define-condition internal-simple-file-error
    (#+CLOS internal-simple-error file-error)
  #-CLOS ((function-name nil) format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%internal-simple-file-error-)
  #-CLOS (:REPORT internal-simple-error-printer)
  #+CLOS ())

(define-condition internal-simple-stream-error 
    (#+CLOS internal-simple-error stream-error)
  #-CLOS ((function-name nil) format-string (format-arguments '()))
  #-CLOS (:CONC-NAME %%internal-simple-stream-error-)
  #-CLOS (:REPORT internal-simple-error-printer)
  #+CLOS ())

#-CLOS
(defun internal-error-function-name (condition)
  (etypecase condition
    (INTERNAL-ERROR                
     (%%internal-error-function-name condition))
    (INTERNAL-SIMPLE-ERROR         
     (%%internal-simple-error-function-name condition))
    (INTERNAL-TYPE-ERROR 
     (%%internal-type-error-function-name condition))
    (INTERNAL-SIMPLE-PROGRAM-ERROR
     (%%internal-simple-program-error-function-name condition))
    (INTERNAL-SIMPLE-CONTROL-ERROR
     (%%internal-simple-control-error-function-name condition))
    (INTERNAL-UNBOUND-VARIABLE  
     (%%internal-unbound-variable-function-name condition))
    (INTERNAL-UNDEFINED-FUNCTION 
     (%%internal-undefined-function-function-name condition))
    (INTERNAL-END-OF-FILE        
     (%%internal-end-of-file-function-name condition))
    (INTERNAL-SIMPLE-FILE-ERROR  
     (%%internal-simple-file-error-function-name condition))
    (INTERNAL-SIMPLE-STREAM-ERROR 
     (%%internal-simple-stream-error-function-name condition))))
)
;;; End new

#-CLOS
(progn

(defun simple-condition-format-string (condition)
  (etypecase condition
    (SIMPLE-CONDITION  (%%simple-condition-format-string  condition))
    (SIMPLE-WARNING    (%%simple-warning-format-string    condition))
    (SIMPLE-TYPE-ERROR (%%simple-type-error-format-string condition))
    (SIMPLE-ERROR      (%%simple-error-format-string      condition))
    (INTERNAL-SIMPLE-ERROR (%%internal-simple-error-format-string condition))
    (INTERNAL-SIMPLE-PROGRAM-ERROR
     (%%internal-simple-program-error-format-string condition))
    (INTERNAL-SIMPLE-CONTROL-ERROR
     (%%internal-simple-control-error-format-string condition))
    (INTERNAL-SIMPLE-FILE-ERROR
     (%%internal-simple-file-error-format-string condition))
    (INTERNAL-SIMPLE-STREAM-ERROR
     (%%internal-simple-stream-error-format-string condition))))

(defun simple-condition-format-arguments (condition)
  (etypecase condition
    (SIMPLE-CONDITION  (%%simple-condition-format-arguments  condition))
    (SIMPLE-WARNING    (%%simple-warning-format-arguments    condition))
    (SIMPLE-TYPE-ERROR (%%simple-type-error-format-arguments condition))
    (SIMPLE-ERROR      (%%simple-error-format-arguments      condition))
    (INTERNAL-SIMPLE-ERROR
     (%%internal-simple-error-format-arguments condition))
    (INTERNAL-SIMPLE-PROGRAM-ERROR
	  (%%internal-simple-program-error-format-arguments condition))
    (INTERNAL-SIMPLE-CONTROL-ERROR
	  (%%internal-simple-control-error-format-arguments condition))
    (INTERNAL-SIMPLE-FILE-ERROR
	  (%%internal-simple-file-error-format-arguments condition))
    (INTERNAL-SIMPLE-STREAM-ERROR
	  (%%internal-simple-stream-error-format-arguments condition))))

(defun simple-condition-class-p (type)
  (member type '(SIMPLE-CONDITION SIMPLE-WARNING SIMPLE-TYPE-ERROR SIMPLE-ERROR
		 INTERNAL-SIMPLE-ERROR
		 INTERNAL-SIMPLE-PROGRAM-ERROR
		 INTERNAL-SIMPLE-CONTROL-ERROR
		 INTERNAL-SIMPLE-FILE-ERROR
		 INTERNAL-SIMPLE-STREAM-ERROR)
	  :TEST #'eq))
)

#+CLOS
(defun simple-condition-class-p (type)
  (typep type 'SIMPLE-CONDITION-CLASS))


(defmacro handler-case (form &rest cases)
  (let ((no-error-clause (assoc ':NO-ERROR cases)))
    (if no-error-clause
	(let ((normal-return (make-symbol "NORMAL-RETURN"))
	      (error-return  (make-symbol "ERROR-RETURN")))
	  `(block ,error-return
	    (multiple-value-call #'(lambda ,@(cdr no-error-clause))
	      (block ,normal-return
		(return-from ,error-return
		  (handler-case (return-from ,normal-return ,form)
		     ,@(remove no-error-clause cases)))))))
	(let ((tag (gensym))
	      (var (gensym))
	      (annotated-cases (mapcar #'(lambda (case) (cons (gensym) case))
				       cases)))
	  `(block ,tag
	     (let ((,var nil))
	       (declare (ignore ,var))
	       (tagbody
		 (handler-bind ,(mapcar #'(lambda (annotated-case)
					    (list (cadr annotated-case)
						  `#'(lambda (temp)
						       ,@(if (caddr annotated-case)
							     `((setq ,var temp)))
						       (go ,(car annotated-case)))))
					annotated-cases)
			       (return-from ,tag ,form))
		 ,@(mapcan #'(lambda (annotated-case)
			       (list (car annotated-case)
				     (let ((body (cdddr annotated-case)))
				       `(return-from ,tag
					  ,(cond ((caddr annotated-case)
						  `(let ((,(caaddr annotated-case)
							  ,var))
						     ,@body))
						 ((not (cdr body))
						  (car body))
						 (t
						  `(progn ,@body)))))))
			   annotated-cases))))))))

(defmacro ignore-errors (&rest forms)
  `(handler-case (progn ,@forms)
     (error (condition) (values nil condition))))

(defun abort          ()      (invoke-restart 'ABORT)
       			      (error 'ABORT-FAILURE))
(defun continue       ()      (invoke-restart 'CONTINUE))
(defun muffle-warning ()      (invoke-restart 'MUFFLE-WARNING))
(defun store-value    (value) (invoke-restart 'STORE-VALUE value))
(defun use-value      (value) (invoke-restart 'USE-VALUE   value))

#|
;;; ----------------------------------------------------------------------
;;; Interface
;;;

(defun sys::universal-error-handler
  (error-name correctable function-name
	      continue-format-string error-format-string
	      args)
  (declare (inline apply)) ;; So as not to get bogus frames in debugger
  (if correctable
      (apply #'cerror continue-format-string args)
      (apply #'error error-format-string args)))

(defun invoke-debugger (condition)
  (let ((si::*tpl-prompt-hook*
	 '(lambda 
  (sys::break-level *restart-clusters*
		    (format nil "~?"
			    (format-string condition)
			    (format-arguments condition))))

(setq *restart-clusters*
      (list (list
	     (make-restart
	      :NAME 'ABORT
	      :FUNCTION #'(lambda () (throw si::*quit-tag* NIL))
	      :REPORT-FUNCTION 
	      #'(lambda (stream) 
		  (format stream "Return to top level."))
	      :INTERACTIVE-FUNCTION nil))))

;;; ----------------------------------------------------------------------
;;; Sample Debugger

(defvar *debug-level* 0)
(defvar *debug-abort* nil)
(defvar *debug-continue* nil)
(defvar *debug-condition* nil)
(defvar *debug-restarts* nil)
(defvar *number-of-debug-restarts* 0)
(defvar *debug-eval* 'eval)
(defvar *debug-print* #'(lambda (values) (format t "~&~{~s~^,~%~}" values)))

(defmacro debug-command                (x) `(get ,x 'debug-command))
(defmacro debug-command-argument-count (x) `(get ,x 'debug-command-argument-count))

(defmacro define-debug-command (name bvl &rest body)
  `(progn (setf (debug-command ',name) #'(lambda ,bvl ,@body))
          (setf (debug-command-argument-count ',name) ,(length bvl))
          ',name))

(defun read-debug-command ()
  (format t "~&debug ~d> " *debug-level*)
  (cond ((char= (peek-char t) #\:)
	 (read-char) ;Eat the ":" so that ":1" reliably reads a number.
	 (with-input-from-string (stream (read-line))
	   (let ((eof (list nil)))
	     (do ((form (let ((*package* (find-package "keyword")))
			  (read stream nil eof))
			(read stream nil eof))
		  (l '() (cons form l)))
		 ((eq form eof) (nreverse l))))))
	(t
	 (list :eval (read)))))
                   
(define-debug-command :EVAL (form)
  (funcall *debug-print* (multiple-value-list (funcall *debug-eval* form))))

(define-debug-command :ABORT ()
  (if *debug-abort*
      (invoke-restart-interactively *debug-abort*)
      (format T "~&There is no way to abort.~%")))

(define-debug-command :CONTINUE ()
  (if *debug-continue*
      (invoke-restart-interactively *debug-continue*)
      (format T "~&There is no way to continue.~%")))

(define-debug-command :ERROR ()
  (format T "~&~A~%" *debug-condition*))

(define-debug-command :HELP ()
  (format T "~&You are in a portable debugger.~
             ~%Type a debugger command or a form to evaluate.~
             ~%Commands are:~%")
  (show-restarts *debug-restarts* *number-of-debug-restarts* 16)
  (format T "~& :EVAL form     Evaluate a form.~
             ~% :HELP          Show this text.~%")
  (when *debug-abort*    (format T "~& :ABORT         Exit by ABORT.~%"))
  (when *debug-continue* (format T "~& :CONTINUE      Exit by CONTINUE.~%"))
  (format T "~& :ERROR         Reprint error message.~%"))



(defvar *debug-command-prefix* ":")

(defun show-restarts (&optional (restarts *debug-restarts*)
		      		(max *number-of-debug-restarts*)
				target-column)
  (unless max (setq max (length restarts)))
  (when restarts
    (do ((w (if target-column
		(- target-column 3)
		(ceiling (log max 10))))
         (p restarts (cdr p))
         (i 0 (1+ i)))
        ((or (not p) (= i max)))
      (format T "~& ~A~A "
	      *debug-command-prefix*
	      (let ((s (format nil "~d" (+ i 1))))
		(with-output-to-string (str)
		  (format str "~A" s)
		  (dotimes (i (- w (length s)))
		    (write-char #\Space str)))))
      (when (eq (car p) *debug-abort*) (format t "(Abort) "))
      (when (eq (car p) *debug-continue*) (format T "(Continue) "))
      (format T "~A" (car p))
      (format T "~%"))))

(defvar *debugger-hook* NIL)

(defun invoke-debugger (&optional (datum "Debug") &rest arguments)
  (let ((condition
	 (coerce-to-condition datum arguments 'simple-condition 'debug)))
    (when *debugger-hook*
      (let ((hook *debugger-hook*)
	    (*debugger-hook* nil))
	(funcall hook condition hook)))
    (standard-debugger condition)))

(defun standard-debugger (condition)
  (let* ((*debug-level* (1+ *debug-level*))
	 (*debug-restarts* (compute-restarts))
	 (*number-of-debug-restarts* (length *debug-restarts*))
	 (*debug-abort*    (find-restart 'abort))
	 (*debug-continue* (or (let ((c (find-restart 'continue)))
				 (if (or (not *debug-continue*)
					 (not (eq *debug-continue* c)))
				     c nil))
			       (let ((c (if *debug-restarts*
					    (first *debug-restarts*) nil)))
				 (if (not (eq c *debug-abort*)) c NIL))))
	 (*debug-condition* condition))
    (format T "~&~A~%" condition)
    (show-restarts)
    (do ((command (read-debug-command) (read-debug-command)))
	(nil)
      (execute-debugger-command (car command) (cdr command) *debug-level*))))

(defun execute-debugger-command (cmd args level)
  (with-simple-restart (abort "Return to debug level ~D." level)
    (when cmd
      (if (integerp cmd)
	  (if (and (plusp cmd)
		   (< cmd (+ *number-of-debug-restarts* 1)))
	      (let ((restart (nth (- cmd 1) *debug-restarts*)))
		(if args
		    (apply #'invoke-restart restart (mapcar *debug-eval* args))
		    (invoke-restart-interactively restart)))
	      (format T "~&No such restart."))
	  (let ((fn (debug-command cmd)))
	    (if fn
		(if (= (length args) (debug-command-argument-count cmd))
		    (apply fn args)
		    (format T "~&Too ~:[few~;many~] arguments to ~A."
			    (> (length args) (debug-command-argument-count cmd))
			    cmd))
		(format T "~&~S is not a debugger command.~%" cmd)))))))

|#
